@@ -59,6 +59,6 @@ extern int modified_cornacchia(mpz_t x, mpz_t y, mpz_t D, mpz_t p);
 extern UV poly_class_poly(IV D, mpz_t**T, int* type);
 
 /* return a 0 terminated list of all D's sorted by degree */
-extern UV* poly_class_degrees(void);
+extern IV* poly_class_degrees(int insert_1s);
 
 #endif
